#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from idb.common.types import TargetDescription


def mock_target(
    udid: str = "udid",
    name: str = "name",
    screen_dimensions: None = None,
    state: str = "state",
    target_type: str = "type",
    os_version: str = "os_version",
    architecture: str = "arch",
    companion_info: None = None,
) -> TargetDescription:
    return TargetDescription(
        udid=udid,
        name=name,
        screen_dimensions=screen_dimensions,
        state=state,
        target_type=target_type,
        os_version=os_version,
        architecture=architecture,
        companion_info=companion_info,
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


from idb.common.companion import CompanionClient
from idb.grpc.idb_pb2 import TargetDescriptionRequest, TargetDescriptionResponse


async def client(client: CompanionClient) -> TargetDescriptionResponse:
    return await client.stub.describe(TargetDescriptionRequest())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List, Optional

from idb.common.companion import CompanionClient
from idb.grpc.idb_pb2 import DebugServerRequest, DebugServerResponse


async def _unary(
    client: CompanionClient, request: DebugServerRequest
) -> DebugServerResponse:
    async with client.stub.debugserver.open() as stream:
        await stream.send_message(request, end=True)
        return await stream.recv_message()


async def debugserver_start(client: CompanionClient, bundle_id: str) -> List[str]:
    response = await _unary(
        client=client,
        request=DebugServerRequest(start=DebugServerRequest.Start(bundle_id=bundle_id)),
    )
    return response.status.lldb_bootstrap_commands


async def debugserver_stop(client: CompanionClient) -> None:
    await _unary(
        client=client, request=DebugServerRequest(stop=DebugServerRequest.Stop())
    )


async def debugserver_status(client: CompanionClient) -> Optional[List[str]]:
    response = await _unary(
        client=client, request=DebugServerRequest(status=DebugServerRequest.Status())
    )
    commands = response.status.lldb_bootstrap_commands
    if not len(commands):
        return None
    return commands


CLIENT_PROPERTIES = [  # pyre-ignore
    debugserver_start,
    debugserver_status,
    debugserver_stop,
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
from argparse import Namespace
from unittest.mock import ANY, MagicMock, patch
from typing import Any

from idb.cli.main import gen_main as cli_main
from idb.common.constants import XCTEST_TIMEOUT
from idb.common.types import CrashLogQuery, HIDButtonType
from idb.utils.testing import AsyncMock, TestCase


class TestParser(TestCase):
    def __init__(self, *args: Any, **kws: Any) -> None:
        super().__init__(*args, loop=asyncio.get_event_loop(), **kws)

    def setUp(self) -> None:
        self.client_mock = MagicMock(name="client_mock")
        self.client_patch = patch("idb.cli.commands.base.IdbClient", self.client_mock)
        self.client_patch.start()
        self

    def tearDown(self) -> None:
        self.client_patch.stop()

    async def test_launch_with_udid(self) -> None:
        bundle_id = "com.foo.app"
        udid = "my udid"
        self.client_mock().launch = AsyncMock(return_value=bundle_id)
        await cli_main(cmd_input=["launch", "--udid", udid, bundle_id])
        self.client_mock().launch.assert_called_once_with(
            bundle_id=bundle_id, env={}, args=[], foreground_if_running=False, stop=None
        )

    async def test_boot(self) -> None:
        self.client_mock().boot = AsyncMock()
        udid = "my udid"
        await cli_main(cmd_input=["boot", "--udid", udid])
        self.client_mock().boot.assert_called_once()

    async def test_install(self) -> None:
        self.client_mock().install = AsyncMock(return_value="com.foo.bar")
        app_path = "testApp.app"
        await cli_main(cmd_input=["install", app_path])
        self.client_mock().install.assert_called_once_with(app_path)

    async def test_uninstall(self) -> None:
        self.client_mock().uninstall = AsyncMock()
        app_path = "com.dummy.app"
        await cli_main(cmd_input=["uninstall", app_path])
        self.client_mock().uninstall.assert_called_once_with(bundle_id=app_path)

    async def test_list_apps(self) -> None:
        self.client_mock().list_apps = AsyncMock(return_value=[])
        await cli_main(cmd_input=["list-apps"])
        self.client_mock().list_apps.assert_called_once()

    async def test_connect_with_host_and_port(self) -> None:
        self.client_mock().connect = AsyncMock()
        host = "someHost"
        port = 1234
        await cli_main(cmd_input=["connect", host, str(port)])
        self.client_mock().connect.assert_called_once_with(
            destination=(host, port, None), metadata=ANY
        )

    async def test_connect_with_host_and_port_and_grpc_port(self) -> None:
        self.client_mock().connect = AsyncMock()
        host = "someHost"
        port = 1234
        grpc_port = 1235
        await cli_main(cmd_input=["connect", host, str(port), str(grpc_port)])
        self.client_mock().connect.assert_called_once_with(
            destination=(host, grpc_port, port), metadata=ANY
        )

    async def test_connect_with_udid(self) -> None:
        self.client_mock().connect = AsyncMock()
        udid = "AAAA-AAAA"
        await cli_main(cmd_input=["connect", udid])
        self.client_mock().connect.assert_called_once_with(
            destination=udid, metadata=ANY
        )

    async def test_disconnect_with_host_and_port(self) -> None:
        self.client_mock().disconnect = AsyncMock()
        host = "someHost"
        port = 1234
        await cli_main(cmd_input=["disconnect", host, str(port)])
        self.client_mock().disconnect.assert_called_once_with(
            destination=(host, port, None)
        )

    async def test_disconnect_with_udid(self) -> None:
        self.client_mock().disconnect = AsyncMock()
        udid = "AAAA-AAAA"
        await cli_main(cmd_input=["disconnect", udid])
        self.client_mock().disconnect.assert_called_once_with(destination=udid)

    async def test_file_mkdir_flag(self) -> None:
        self.client_mock().mkdir = AsyncMock()
        src = "path"
        bundle_id = "com.bundle.id"
        cmd_input = ["file", "mkdir", src, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().mkdir.assert_called_once_with(bundle_id=bundle_id, path=src)

    async def test_file_mkdir_colon(self) -> None:
        self.client_mock().mkdir = AsyncMock()
        cmd_input = ["file", "mkdir", "com.bundle.id:path"]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().mkdir.assert_called_once_with(
            bundle_id="com.bundle.id", path="path"
        )

    async def test_file_rmpath_flag(self) -> None:
        self.client_mock().rm = AsyncMock()
        src = "path"
        bundle_id = "com.bundle.id"
        cmd_input = ["file", "remove", src, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().rm.assert_called_once_with(bundle_id=bundle_id, paths=[src])

    async def test_file_rmpath_flag_multiple(self) -> None:
        self.client_mock().rm = AsyncMock()
        srcs = ["pathA", "pathB"]
        bundle_id = "com.bundle.id"
        cmd_input = ["file", "remove", srcs[0], srcs[1], "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().rm.assert_called_once_with(bundle_id=bundle_id, paths=srcs)

    async def test_file_rmpath_colon(self) -> None:
        self.client_mock().rm = AsyncMock()
        cmd_input = ["file", "remove", "com.bundle.id:path"]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().rm.assert_called_once_with(
            bundle_id="com.bundle.id", paths=["path"]
        )

    async def test_file_listpath_flag(self) -> None:
        self.client_mock().ls = AsyncMock(return_value=[])
        src = "path"
        bundle_id = "com.bundle.id"
        cmd_input = ["file", "list", src, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().ls.assert_called_once_with(bundle_id=bundle_id, path=src)

    async def test_file_listpath_colon(self) -> None:
        self.client_mock().ls = AsyncMock(return_value=[])
        cmd_input = ["file", "list", "com.bundle.id:path"]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().ls.assert_called_once_with(
            bundle_id="com.bundle.id", path="path"
        )

    async def test_file_move_flag(self) -> None:
        self.client_mock().mv = AsyncMock(return_value=[])
        src = "a"
        dst = "b"
        bundle_id = "com.bundle.id"
        cmd_input = ["file", "move", src, dst, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().mv.assert_called_once_with(
            bundle_id=bundle_id, src_paths=[src], dest_path=dst
        )

    async def test_file_move_flag_multiple(self) -> None:
        self.client_mock().mv = AsyncMock(return_value=[])
        srcs = ["src1", "src2"]
        dst = "b"
        bundle_id = "com.bundle.id"
        cmd_input = ["file", "move", srcs[0], srcs[1], dst, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().mv.assert_called_once_with(
            bundle_id=bundle_id, src_paths=srcs, dest_path=dst
        )

    async def test_file_move_colon(self) -> None:
        self.client_mock().mv = AsyncMock(return_value=[])
        src = "a"
        cmd_input = ["file", "move", src, "com.bundle.id:b"]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().mv.assert_called_once_with(
            bundle_id="com.bundle.id", src_paths=[src], dest_path="b"
        )

    async def test_file_push_single_flag(self) -> None:
        self.client_mock().push = AsyncMock()
        bundle_id = "com.myapp"
        src = "Library/myFile.txt"
        dst = "someOutputDir"
        cmd_input = ["file", "push", src, dst, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().push.assert_called_once_with(
            bundle_id=bundle_id, src_paths=[src], dest_path=dst
        )

    async def test_file_push_multi_flag(self) -> None:
        self.client_mock().push = AsyncMock(return_value=[])
        bundle_id = "com.myapp"
        src1 = "Library/myFile.txt"
        src2 = "Library/myFile2.txt"
        dst = "someOutputDir"
        cmd_input = ["file", "push", src1, src2, dst, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().push.assert_called_once_with(
            bundle_id=bundle_id, src_paths=[src1, src2], dest_path=dst
        )

    async def test_file_push_single_colon(self) -> None:
        self.client_mock().push = AsyncMock(return_value=[])
        src = "Library/myFile.txt"
        cmd_input = ["file", "push", src, "com.bundle.id:someOutputDir"]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().push.assert_called_once_with(
            bundle_id="com.bundle.id", src_paths=[src], dest_path="someOutputDir"
        )

    async def test_file_push_multi_colon(self) -> None:
        self.client_mock().push = AsyncMock(return_value=[])
        src1 = "Library/myFile.txt"
        src2 = "Library/myFile2.txt"
        cmd_input = ["file", "push", src1, src2, "com.bundle.id:someOutputDir"]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().push.assert_called_once_with(
            bundle_id="com.bundle.id", src_paths=[src1, src2], dest_path="someOutputDir"
        )

    async def test_file_pull(self) -> None:
        self.client_mock().pull = AsyncMock(return_value=[])
        bundle_id = "com.myapp"
        src = "Library/myFile.txt"
        dst = "someOutputDir"
        cmd_input = ["file", "pull", src, dst, "--bundle-id", bundle_id]
        await cli_main(cmd_input=cmd_input)
        self.client_mock().pull.assert_called_once_with(
            bundle_id=bundle_id, src_path=src, dest_path=dst
        )

    async def test_push_deprecated(self) -> None:
        self.client_mock().push = AsyncMock(return_value=[])
        src_path = "./fileToSend.txt"
        app_bundle_id = "com.myapp"
        dest_path = "Library"
        await cli_main(cmd_input=["push", src_path, app_bundle_id, dest_path])
        self.client_mock().push.assert_called_once_with(
            bundle_id=app_bundle_id, src_paths=[src_path], dest_path=dest_path
        )

    async def test_pull_deprecated(self) -> None:
        self.client_mock().pull = AsyncMock(return_value=[])
        bundle_id = "com.myapp"
        src = "Library/myFile.txt"
        dest = "someOutputDir"
        await cli_main(cmd_input=["pull", bundle_id, src, dest])
        self.client_mock().pull.assert_called_once_with(
            bundle_id=bundle_id, src_path=src, dest_path=dest
        )

    async def test_list_targets(self) -> None:
        self.client_mock().list_targets = AsyncMock(return_value=[])
        await cli_main(cmd_input=["list-targets"])
        self.client_mock().list_targets.assert_called_once()

    async def test_kill(self) -> None:
        mock = AsyncMock()
        with patch("idb.cli.commands.kill.IdbClient.kill", new=mock, create=True):
            await cli_main(cmd_input=["kill"])
            mock.assert_called_once()

    async def test_xctest_install(self) -> None:
        self.client_mock().install_xctest = AsyncMock(return_value=[])
        test_bundle_path = "testBundle.xctest"
        await cli_main(cmd_input=["xctest", "install", test_bundle_path])
        self.client_mock().install_xctest.assert_called_once_with(test_bundle_path)

    def xctest_run_namespace(self, command: str, test_bundle_id: str) -> Namespace:
        namespace = Namespace()
        namespace.log_level = "INFO"
        namespace.root_command = "xctest"
        namespace.xctest = "run"
        namespace.udid = None
        namespace.json = False
        if command in ["app", "ui"]:
            namespace.tests_to_run = None
            namespace.tests_to_skip = None
        elif command == "logic":
            namespace.test_to_run = None
        namespace.run = command
        namespace.test_bundle_id = test_bundle_id
        namespace.test_arguments = []
        namespace.daemon_port = 9888
        namespace.daemon_grpc_port = 9889
        namespace.daemon_host = "localhost"
        namespace.result_bundle_path = None
        return namespace

    async def test_xctest_run_app(self) -> None:
        mock = AsyncMock()
        mock.return_value = []
        with patch(
            "idb.cli.commands.xctest_run.RunXcTestAppCommand.run", new=mock, create=True
        ):
            test_bundle_id = "com.me.tests"
            app_under_test_id = "com.me.app"
            await cli_main(
                cmd_input=["xctest", "run", "app", test_bundle_id, app_under_test_id]
            )
            namespace = self.xctest_run_namespace("app", test_bundle_id)
            namespace.app_bundle_id = app_under_test_id
            namespace.force = False
            namespace.timeout = XCTEST_TIMEOUT
            namespace.grpc = False
            mock.assert_called_once_with(namespace, Namespace())

    async def test_xctest_run_ui(self) -> None:
        mock = AsyncMock()
        mock.return_value = []
        with patch(
            "idb.cli.commands.xctest_run.RunXcTestUICommand.run", new=mock, create=True
        ):
            test_bundle_id = "com.me.tests"
            app_under_test_id = "com.me.app"
            test_host_app_bundle_id = "com.host.app"
            await cli_main(
                cmd_input=[
                    "xctest",
                    "run",
                    "ui",
                    test_bundle_id,
                    app_under_test_id,
                    test_host_app_bundle_id,
                ]
            )
            namespace = self.xctest_run_namespace("ui", test_bundle_id)
            namespace.app_bundle_id = app_under_test_id
            namespace.test_host_app_bundle_id = test_host_app_bundle_id
            namespace.force = False
            namespace.timeout = XCTEST_TIMEOUT
            namespace.grpc = False
            mock.assert_called_once_with(namespace, Namespace())

    async def test_xctest_run_logic(self) -> None:
        mock = AsyncMock()
        mock.return_value = []
        with patch(
            "idb.cli.commands.xctest_run.CommonRunXcTestCommand.run",
            new=mock,
            create=True,
        ):
            test_bundle_id = "com.me.tests"
            await cli_main(cmd_input=["xctest", "run", "logic", test_bundle_id])
            namespace = self.xctest_run_namespace("logic", test_bundle_id)
            namespace.timeout = XCTEST_TIMEOUT
            namespace.force = False
            namespace.grpc = False
            mock.assert_called_once_with(namespace, Namespace())

    async def test_xctest_list(self) -> None:
        self.client_mock().list_xctests = AsyncMock(return_value=[])
        await cli_main(cmd_input=["xctest", "list"])
        self.client_mock().list_xctests.assert_called_once()

    async def test_xctest_list_bundles(self) -> None:
        self.client_mock().list_test_bundle = AsyncMock(return_value=[])
        bundle_id = "myBundleID"
        await cli_main(cmd_input=["xctest", "list-bundle", bundle_id])
        self.client_mock().list_test_bundle.assert_called_once_with(
            test_bundle_id=bundle_id
        )

    async def test_daemon(self) -> None:
        mock = AsyncMock()
        with patch(
            "idb.cli.commands.daemon.DaemonCommand._run_impl", new=mock, create=True
        ):
            port = 1234
            grpc_port = 1235
            await cli_main(
                cmd_input=[
                    "daemon",
                    "--daemon-port",
                    str(port),
                    "--daemon-grpc-port",
                    str(grpc_port),
                ]
            )
            namespace = Namespace()
            namespace.daemon_port = port
            namespace.daemon_grpc_port = grpc_port
            namespace.log_level = "INFO"
            namespace.root_command = "daemon"
            namespace.json = False
            namespace.reply_fd = None
            namespace.prefer_ipv6 = False
            namespace.notifier_path = None
            mock.assert_called_once_with(namespace)

    async def test_terminate(self) -> None:
        self.client_mock().terminate = AsyncMock(return_value=[])
        bundle_id = "com.foo.app"
        udid = "my udid"
        await cli_main(cmd_input=["terminate", bundle_id, "--udid", udid])
        self.client_mock().terminate.assert_called_once_with(bundle_id)

    async def test_log(self) -> None:
        mock = AsyncMock()
        with patch("idb.cli.commands.log.LogCommand._run_impl", new=mock, create=True):
            await cli_main(cmd_input=["log", "--udid", "1234"])
            namespace = Namespace()
            namespace.log_level = "INFO"
            namespace.root_command = "log"
            namespace.udid = "1234"
            namespace.json = False
            namespace.log_arguments = []
            namespace.daemon_port = 9888
            namespace.daemon_grpc_port = 9889
            namespace.daemon_host = "localhost"
            namespace.force = False
            namespace.grpc = False
            mock.assert_called_once_with(namespace)

    async def test_log_arguments(self) -> None:
        mock = AsyncMock()
        with patch("idb.cli.commands.log.LogCommand._run_impl", new=mock, create=True):
            await cli_main(cmd_input=["log", "--", "--style", "json"])
            namespace = Namespace()
            namespace.log_level = "INFO"
            namespace.root_command = "log"
            namespace.udid = None
            namespace.json = False
            namespace.log_arguments = ["--", "--style", "json"]
            namespace.daemon_port = 9888
            namespace.daemon_grpc_port = 9889
            namespace.daemon_host = "localhost"
            namespace.force = False
            namespace.grpc = False
            mock.assert_called_once_with(namespace)

    async def test_clear_keychain(self) -> None:
        self.client_mock().clear_keychain = AsyncMock(return_value=[])
        await cli_main(cmd_input=["clear-keychain"])
        self.client_mock().clear_keychain.assert_called_once()

    async def test_open_url(self) -> None:
        self.client_mock().open_url = AsyncMock(return_value=[])
        url = "http://facebook.com"
        await cli_main(cmd_input=["open", url])
        self.client_mock().open_url.assert_called_once_with(url)

    async def test_set_location(self) -> None:
        self.client_mock().set_location = AsyncMock(return_value=[])
        latitude = 1.0
        longitude = 2.0
        await cli_main(cmd_input=["set-location", str(latitude), str(longitude)])
        self.client_mock().set_location.assert_called_once_with(latitude, longitude)

    async def test_approve(self) -> None:
        self.client_mock().approve = AsyncMock(return_value=[])
        bundle_id = "com.fb.myApp"
        await cli_main(cmd_input=["approve", bundle_id, "photos"])
        self.client_mock().approve.assert_called_once_with(
            bundle_id=bundle_id, permissions={"photos"}
        )

    async def test_video(self) -> None:
        mock = AsyncMock()
        with patch(
            "idb.cli.commands.record.RecordVideoCommand._run_impl",
            new=mock,
            create=True,
        ):
            output_file = "video.mp4"
            await cli_main(cmd_input=["record-video", output_file])
            namespace = Namespace()
            namespace.log_level = "INFO"
            namespace.root_command = "record-video"
            namespace.udid = None
            namespace.json = False
            namespace.output_file = output_file
            namespace.daemon_port = 9888
            namespace.daemon_grpc_port = 9889
            namespace.daemon_host = "localhost"
            namespace.force = False
            namespace.grpc = False
            mock.assert_called_once_with(namespace)

    async def test_key_sequence(self) -> None:
        self.client_mock().key_sequence = AsyncMock(return_value=[])
        await cli_main(cmd_input=["ui", "key-sequence", "1", "2", "3"])
        self.client_mock().key_sequence.assert_called_once_with(key_sequence=[1, 2, 3])

    async def test_tap(self) -> None:
        self.client_mock().tap = AsyncMock(return_value=[])
        await cli_main(cmd_input=["ui", "tap", "10", "20"])
        self.client_mock().tap.assert_called_once_with(x=10, y=20, duration=None)

    async def test_button(self) -> None:
        self.client_mock().button = AsyncMock(return_value=[])
        await cli_main(cmd_input=["ui", "button", "SIRI"])
        self.client_mock().button.assert_called_once_with(
            button_type=HIDButtonType.SIRI, duration=None
        )

    async def test_key(self) -> None:
        self.client_mock().key = AsyncMock(return_value=[])
        await cli_main(cmd_input=["ui", "key", "12"])
        self.client_mock().key.assert_called_once_with(keycode=12, duration=None)

    async def test_text_input(self) -> None:
        self.client_mock().text = AsyncMock(return_value=[])
        text = "Some Text"
        await cli_main(cmd_input=["ui", "text", text])
        self.client_mock().text.assert_called_once_with(text=text)

    async def test_swipe_with_delta(self) -> None:
        self.client_mock().swipe = AsyncMock(return_value=[])
        await cli_main(cmd_input=["ui", "swipe", "1", "2", "3", "4", "--delta", "5"])
        self.client_mock().swipe.assert_called_once_with(
            p_start=(1, 2), p_end=(3, 4), delta=5
        )

    async def test_swipe_without_delta(self) -> None:
        self.client_mock().swipe = AsyncMock(return_value=[])
        await cli_main(cmd_input=["ui", "swipe", "1", "2", "3", "4"])
        self.client_mock().swipe.assert_called_once_with(
            p_start=(1, 2), p_end=(3, 4), delta=None
        )

    async def test_contacts_update(self) -> None:
        self.client_mock().contacts_update = AsyncMock(return_value=[])
        await cli_main(cmd_input=["contacts", "update", "/dev/null"])
        self.client_mock().contacts_update.assert_called_once_with(
            contacts_path="/dev/null"
        )

    async def test_accessibility_info_all(self) -> None:
        self.client_mock().accessibility_info = AsyncMock()
        await cli_main(cmd_input=["ui", "describe-all"])
        self.client_mock().accessibility_info.assert_called_once_with(point=None)

    async def test_accessibility_info_at_point(self) -> None:
        self.client_mock().accessibility_info = AsyncMock()
        await cli_main(cmd_input=["ui", "describe-point", "10", "20"])
        self.client_mock().accessibility_info.assert_called_once_with(point=(10, 20))

    async def test_crash_list_all(self) -> None:
        self.client_mock().crash_list = AsyncMock(return_value=[])
        await cli_main(cmd_input=["crash", "list"])
        self.client_mock().crash_list.assert_called_once_with(query=CrashLogQuery())

    async def test_crash_list_with_predicate(self) -> None:
        self.client_mock().crash_list = AsyncMock(return_value=[])
        await cli_main(cmd_input=["crash", "list", "--since", "20"])
        self.client_mock().crash_list.assert_called_once_with(
            query=CrashLogQuery(since=20)
        )

    async def test_crash_show(self) -> None:
        self.client_mock().crash_show = AsyncMock()
        await cli_main(cmd_input=["crash", "show", "foo"])
        self.client_mock().crash_show.assert_called_once_with(name="foo")

    async def test_crash_delete_all(self) -> None:
        self.client_mock().crash_delete = AsyncMock(return_value=[])
        await cli_main(cmd_input=["crash", "delete", "--all"])
        self.client_mock().crash_delete.assert_called_once_with(query=CrashLogQuery())

    async def test_crash_delete_with_predicate(self) -> None:
        self.client_mock().crash_delete = AsyncMock(return_value=[])
        await cli_main(cmd_input=["crash", "delete", "--since", "20"])
        self.client_mock().crash_delete.assert_called_once_with(
            query=CrashLogQuery(since=20)
        )

    async def test_crash_delete_with_name(self) -> None:
        self.client_mock().crash_delete = AsyncMock(return_value=[])
        await cli_main(cmd_input=["crash", "delete", "some.foo.bar.crash"])
        self.client_mock().crash_delete.assert_called_once_with(
            query=CrashLogQuery(name="some.foo.bar.crash")
        )

    async def test_instruments(self) -> None:
        self.client_mock().run_instruments = AsyncMock()
        template_name = "System Trace"
        trace_path = template_name + ".trace"
        await cli_main(cmd_input=["instruments", template_name])
        self.client_mock().run_instruments.assert_called_once_with(
            stop=ANY,
            template=template_name,
            app_bundle_id=None,
            post_process_arguments=None,
            trace_path=trace_path,
            env={},
            app_args=None,
        )

    async def test_instruments_with_post_args(self) -> None:
        self.client_mock().run_instruments = AsyncMock()
        template_name = "Time Profiler"
        trace_path = template_name + ".trace"
        post_process_arguments = ["instrumental", "convert"]
        await cli_main(
            cmd_input=[
                "instruments",
                template_name,
                "--post-args",
                *post_process_arguments,
            ]
        )
        self.client_mock().run_instruments.assert_called_once_with(
            stop=ANY,
            template=template_name,
            app_bundle_id=None,
            post_process_arguments=post_process_arguments,
            trace_path=trace_path,
            env={},
            app_args=None,
        )

    async def test_instruments_with_app_args(self) -> None:
        self.client_mock().run_instruments = AsyncMock()
        template_name = "System Trace"
        trace_path = template_name + ".trace"
        await cli_main(
            cmd_input=["instruments", template_name, "--app-args", "perfLab"]
        )
        self.client_mock().run_instruments.assert_called_once_with(
            stop=ANY,
            template=template_name,
            app_bundle_id=None,
            post_process_arguments=None,
            trace_path=trace_path,
            env={},
            app_args=["perfLab"],
        )

    async def test_add_media(self) -> None:
        self.client_mock().add_media = AsyncMock(return_value=["aaa", "bbb"])
        file_paths = ["cat.jpeg", "dog.mov"]
        await cli_main(cmd_input=["add-media"] + file_paths)
        self.client_mock().add_media.assert_called_once_with(file_paths=file_paths)

    async def test_focus(self) -> None:
        self.client_mock().focus = AsyncMock(return_value=["aaa", "bbb"])
        await cli_main(cmd_input=["focus"])
        self.client_mock().focus.assert_called_once()

    async def test_debugserver_start(self) -> None:
        self.client_mock().debugserver_start = AsyncMock(return_value=["aaa", "bbb"])
        await cli_main(cmd_input=["debugserver", "start", "com.foo.bar"])
        self.client_mock().debugserver_start.assert_called_once_with(
            bundle_id="com.foo.bar"
        )

    async def test_debugserver_stop(self) -> None:
        self.client_mock().debugserver_stop = AsyncMock(return_value=["aaa", "bbb"])
        await cli_main(cmd_input=["debugserver", "stop"])
        self.client_mock().debugserver_stop.assert_called_once()

    async def test_debugserver_status(self) -> None:
        self.client_mock().debugserver_status = AsyncMock(return_value=["aaa", "bbb"])
        await cli_main(cmd_input=["debugserver", "status"])
        self.client_mock().debugserver_status.assert_called_once()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
import logging
from typing import List, Optional


class BootManager:
    def __init__(self, companion_path: Optional[str]) -> None:
        self.companion_path = companion_path

    async def boot(self, udid: str) -> None:
        if self.companion_path:
            cmd: List[str] = [self.companion_path, "--boot", udid]
            process = await asyncio.create_subprocess_exec(
                *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            await process.communicate()
        else:
            logging.error(
                "Booting requires the daemon to be started with --notifier-path"
            )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import logging
import warnings
from typing import Optional, Dict

import idb.grpc.ipc_loader as ipc_loader
from idb.common.boot_manager import BootManager
from idb.common.companion import CompanionClient
from idb.manager.companion import CompanionManager
from idb.grpc.ipc_loader import DaemonContext
from idb.grpc.idb_grpc import CompanionServiceBase


# Don't let the abstractmetod machineary mess raise at runtime
CompanionServiceBase.__abstractmethods__ = frozenset([])
# this is to silence the channel not closed warning
# https://github.com/vmagamedov/grpclib/issues/58
warnings.filterwarnings(action="ignore", category=ResourceWarning)


class GRPCHandler(CompanionServiceBase):
    def __init__(
        self,
        companion_manager: CompanionManager,
        boot_manager: BootManager,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        self.logger: logging.Logger = (
            logger if logger else logging.getLogger("idb_daemon")
        )
        self.companion_manager = companion_manager
        self.boot_manager = boot_manager
        for (call_name, f) in ipc_loader.daemon_calls(
            companion_provider=self.provide_client,
            context_provider=self.provide_context,
        ):
            setattr(self, call_name, f)

    def get_udid(self, metadata: Dict[str, str]) -> Optional[str]:
        return metadata.get("udid")

    async def provide_client(self, udid: Optional[str]) -> CompanionClient:
        return await self.companion_manager.get_stub_for_udid(udid=udid)

    async def provide_context(self) -> DaemonContext:
        return DaemonContext(
            companion_manager=self.companion_manager, boot_manager=self.boot_manager
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import json
import os
from argparse import ArgumentParser, Namespace
from typing import Optional, Dict

from idb.common.signal import signal_handler_event
from idb.cli.commands.base import BaseCommand
from idb.client.daemon_pid_saver import remove_daemon_pid, save_daemon_pid
from idb.common.constants import DEFAULT_DAEMON_GRPC_PORT, DEFAULT_DAEMON_PORT
from idb.daemon.server import start_daemon_server
from idb.common.types import Server, IdbException


class DaemonCommand(BaseCommand):
    @property
    def description(self) -> str:
        return "Start the daemon"

    @property
    def name(self) -> str:
        return "daemon"

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--daemon-port",
            help="Port for the daemon to listen on",
            type=int,
            default=DEFAULT_DAEMON_PORT,
        )
        parser.add_argument(
            "--reply-fd",
            help="File descriptor to write port the daemon was started on",
            type=int,
            required=False,
        )
        parser.add_argument(
            "--notifier-path",
            help="path to binary that notifies daemon about devices/simulators.",
            type=str,
            required=False,
        )
        parser.add_argument(
            "--daemon-grpc-port",
            help="Port for the daemon to listen to grpc on",
            type=int,
            default=DEFAULT_DAEMON_GRPC_PORT,
        )
        parser.add_argument(
            "--prefer-ipv6",
            help="If set, always return ipv6 port if available when --reply-fd is used",
            action="store_true",
            default=False,
        )
        super().add_parser_arguments(parser)

    async def _run_impl(self, args: Namespace) -> None:
        server: Optional[Server] = None
        try:
            server = await start_daemon_server(args=args, logger=self.logger)
            save_daemon_pid(pid=os.getpid())
            print(json.dumps(server.ports), sep="\n", flush=True)
            self._reply_with_port(args.reply_fd, args.prefer_ipv6, server.ports)
            await signal_handler_event("server").wait()
        except IdbException as ex:
            self.logger.exception("Exception in main")
            raise ex
        finally:
            remove_daemon_pid(pid=os.getpid())
            if server:
                server.close()
                await server.wait_closed()
            self.logger.info("Exiting")

    def _reply_with_port(
        self, reply_fd: Optional[int], prefer_ipv6: bool, ports: Dict[str, int]
    ) -> None:
        if not reply_fd:
            return

        prefix = "ipv6_" if prefer_ipv6 else "ipv4_"
        all_ports = {
            key.split(prefix)[1]: value
            for (key, value) in ports.items()
            if key.startswith(prefix)
        }
        self.logger.info(f"Replying to fd {reply_fd} with {all_ports}")
        with os.fdopen(reply_fd, "w") as f:
            f.write(json.dumps(all_ports) + "\n")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from idb.common.companion import CompanionClient
from idb.grpc.idb_pb2 import PullRequest, PullResponse, Payload
from idb.grpc.stream import generate_bytes

from idb.common.tar import drain_untar


async def client(
    client: CompanionClient, bundle_id: str, src_path: str, dest_path: str
) -> None:
    await client.stub.pull(
        PullRequest(bundle_id=bundle_id, src_path=src_path, dst_path=dest_path)
    )


async def daemon(client: CompanionClient, request: PullRequest) -> PullResponse:
    destination = request.dst_path
    async with client.stub.pull.open() as stream:
        if not client.is_local:
            # not sending the destination to remote companion
            # so it streams the file back
            request = PullRequest(
                bundle_id=request.bundle_id, src_path=request.src_path, dst_path=None
            )
        await stream.send_message(request, end=True)
        if client.is_local:
            await stream.recv_message()
        else:
            await drain_untar(generate_bytes(stream), output_path=destination)
        client.logger.info(f"pulled file to {destination}")
    return PullResponse(payload=Payload(file_path=destination))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from argparse import ArgumentParser, Namespace

from idb.cli.commands.base import TargetCommand
from idb.client.client import IdbClient


class AddMediaCommand(TargetCommand):
    @property
    def description(self) -> str:
        return "Add photos/videos to the target"

    @property
    def name(self) -> str:
        return "add-media"

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "file_paths", nargs="+", help="Paths to all media files to add"
        )
        super().add_parser_arguments(parser)

    async def run_with_client(self, args: Namespace, client: IdbClient) -> None:
        await client.add_media(file_paths=args.file_paths)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from argparse import ArgumentParser, Namespace


from idb.cli.commands.base import TargetCommand
from idb.client.client import IdbClient


class UninstallCommand(TargetCommand):
    @property
    def description(self) -> str:
        return "Uninstall an application"

    @property
    def name(self) -> str:
        return "uninstall"

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "bundle_id", help="Bundle ID of application to uninstall", type=str
        )
        super().add_parser_arguments(parser)

    async def run_with_client(self, args: Namespace, client: IdbClient) -> None:
        await client.uninstall(bundle_id=args.bundle_id)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


import asyncio
from typing import AsyncIterator

from idb.common.companion import CompanionClient
from idb.grpc.stream import Stream
from idb.common.gzip import drain_gzip_decompress
from idb.grpc.idb_pb2 import Payload, RecordRequest, RecordResponse
from idb.utils.typing import none_throws


Start = RecordRequest.Start
Stop = RecordRequest.Stop


async def _generate_bytes(
    stream: AsyncIterator[RecordResponse],
) -> AsyncIterator[bytes]:
    async for response in stream:
        data = response.payload.data
        yield data


async def daemon(
    client: CompanionClient, stream: Stream[RecordResponse, RecordRequest]
) -> None:
    client.logger.info(f"Starting connection to backend")
    request = await stream.recv_message()
    output_file = none_throws(request).start.file_path
    async with client.stub.record.open() as forward_stream:
        if client.is_local:
            client.logger.info(f"Starting video recording to local file {output_file}")
            await forward_stream.send_message(
                RecordRequest(start=Start(file_path=output_file))
            )
        else:
            client.logger.info(f"Starting video recording with response data")
            await forward_stream.send_message(
                RecordRequest(start=Start(file_path=None))
            )
        client.logger.info("Request sent")
        await stream.recv_message()
        client.logger.info("Stopping video recording")
        await forward_stream.send_message(RecordRequest(stop=Stop()), end=True)
        if client.is_local:
            client.logger.info("Responding with file path")
            response = await forward_stream.recv_message()
            await stream.send_message(response)
        else:
            client.logger.info(f"Decompressing gzip to {output_file}")
            await drain_gzip_decompress(
                _generate_bytes(forward_stream), output_path=output_file
            )
            client.logger.info(f"Finished decompression to {output_file}")
            await stream.send_message(
                RecordResponse(payload=Payload(file_path=output_file))
            )


async def record_video(
    client: CompanionClient, stop: asyncio.Event, output_file: str
) -> None:
    client.logger.info(f"Starting connection to backend")
    async with client.stub.record.open() as stream:
        client.logger.info("Starting video recording")
        await stream.send_message(RecordRequest(start=Start(file_path=output_file)))
        client.logger.info("Request sent")
        await stop.wait()
        client.logger.info("Stopping video recording")
        await stream.send_message(RecordRequest(stop=Stop()), end=True)
        await stream.recv_message()


CLIENT_PROPERTIES = [record_video]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from idb.common.types import CrashLog
from idb.common.companion import CompanionClient
from idb.grpc.idb_pb2 import CrashShowRequest
from idb.ipc.mapping.crash import _to_crash_log


async def client(client: CompanionClient, name: str) -> CrashLog:
    response = await client.stub.crash_show(CrashShowRequest(name=name))
    return _to_crash_log(response)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
import tempfile
from logging import Logger
from typing import AsyncContextManager, Dict, List, Optional

from grpclib.client import Channel
from idb.common.companion import CompanionClient
from idb.common.companion_spawner import CompanionSpawner
from idb.common.types import (
    Address,
    CompanionInfo,
    ConnectionDestination,
    TargetDescription,
)
from idb.grpc.idb_grpc import CompanionServiceStub
from idb.grpc.idb_pb2 import ConnectRequest
from idb.utils.contextlib import asynccontextmanager
from idb.utils.typing import none_throws


class CompanionManager:
    def __init__(self, companion_path: Optional[str], logger: Logger) -> None:
        self._udid_companion_map: Dict[str, CompanionInfo] = {}
        self._udid_target_map: Dict[str, TargetDescription] = {}
        self.companion_spawner = (
            CompanionSpawner(companion_path=companion_path) if companion_path else None
        )
        self._stub_map: Dict[str, CompanionServiceStub] = {}
        self._logger = logger

    def close(self) -> None:
        self._logger.info("Stopping companion manager")
        if self.companion_spawner:
            self.companion_spawner.close()

    def is_companion_available_for_target_udid(self, target_udid: str) -> bool:
        return target_udid in self._udid_target_map

    def add_companion(self, companion: CompanionInfo) -> None:
        self._udid_companion_map[companion.udid] = companion
        self._logger.info(f"Added a companion: {companion}")
        if companion.udid in self._udid_target_map:
            target = self._udid_target_map[companion.udid]
            self._udid_target_map[companion.udid] = TargetDescription(
                udid=target.udid,
                name=target.name,
                state=target.state,
                target_type=target.target_type,
                os_version=target.os_version,
                architecture=target.architecture,
                companion_info=companion,
                screen_dimensions=target.screen_dimensions,
            )
            self._logger.info(f"Assigned the companion to target: {companion.udid}")
        else:
            self._udid_target_map[companion.udid] = TargetDescription(
                udid=companion.udid,
                name="",
                state=None,
                target_type=None,
                os_version=None,
                architecture=None,
                companion_info=companion,
                screen_dimensions=None,
            )

    def remove_companion(
        self, destination: ConnectionDestination
    ) -> Optional[CompanionInfo]:
        self._logger.info(f"Removing companion {destination}")
        if isinstance(destination, Address):
            host = destination.host
            port = destination.port
            for companion in self._udid_companion_map.values():
                if companion.port == port and companion.host == host:
                    del self._udid_companion_map[companion.udid]
                    del self._udid_target_map[companion.udid]
                    self._logger.info(
                        f"Removed a companion at host: {host} port: {port}"
                    )
                    return companion
            self._logger.error(f"Did not found a companion at host {host} port: {port}")
        elif isinstance(destination, str):
            target_udid = destination
            if target_udid in self._udid_companion_map:
                companion = self._udid_companion_map[target_udid]
                del self._udid_companion_map[target_udid]
                del self._udid_target_map[target_udid]
                self._logger.info(f"Removed a companion: {target_udid}")
                return companion
            else:
                self._logger.error(f"Did not find a companion: {target_udid}")
        return None

    @property
    def targets(self) -> List[TargetDescription]:
        return [
            TargetDescription(
                udid=target.udid,
                state=target.state,
                target_type=target.target_type,
                name=target.name or "",
                os_version=target.os_version,
                architecture=target.architecture,
                companion_info=target.companion_info,
                screen_dimensions=None,
            )
            for target in self._udid_target_map.values()
        ]

    def update_target(self, target: TargetDescription) -> None:
        self._udid_target_map[target.udid] = target

    def has_default_companion(self) -> bool:
        return len(self._udid_companion_map) == 1

    def get_default_companion(self) -> CompanionInfo:
        return next(iter(self._udid_companion_map.values()))

    def _get_companion_for_target(
        self, target_udid: Optional[str], timeout: Optional[float] = None
    ) -> Optional[CompanionInfo]:
        self._logger.debug(f"fetching companion for {target_udid}")
        if target_udid and target_udid in self._udid_companion_map:
            return self._udid_companion_map[target_udid]
        elif self.has_default_companion():
            self._logger.info("using default companion")
            return self.get_default_companion()
        self._logger.debug(f"no companion available for {target_udid}")
        return None

    @asynccontextmanager  # noqa T484
    async def create_companion_for_target_with_destination(
        self,
        destination: ConnectionDestination,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
    ) -> AsyncContextManager[CompanionInfo]:
        if isinstance(destination, Address):
            async with self.create_companion_for_target_with_address(
                address=destination, metadata=metadata, timeout=timeout
            ) as companion:
                yield companion
        else:
            async with self.create_companion_for_target_with_udid(
                target_udid=destination, metadata=metadata, timeout=timeout
            ) as companion:
                yield companion

    @asynccontextmanager  # noqa T484
    async def create_companion_for_target_with_address(
        self,
        address: Address,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
    ) -> AsyncContextManager[CompanionInfo]:
        try:
            yield await self._get_companion_info(address, metadata, timeout)
        except OSError as error:
            raise Exception(
                f"Failed to connect to companion at "
                f"{address.host}:{address.port} found error {error}"
            )

    async def _get_companion_info(
        self,
        address: Address,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
    ) -> CompanionInfo:
        stub = self.get_stub_for_address(address.host, none_throws(address.grpc_port))
        with tempfile.NamedTemporaryFile(mode="w+b") as file:
            response = await stub.connect(
                ConnectRequest(metadata=metadata, local_file_path=file.name)
            )
        info = CompanionInfo(
            udid=response.companion.udid,
            host=address.host,
            port=address.port,
            is_local=response.companion.is_local,
            grpc_port=address.grpc_port,
        )
        self.add_companion(info)
        return info

    @asynccontextmanager  # noqa T484
    async def create_companion_for_target_with_udid(
        self,
        target_udid: Optional[str],
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
    ) -> AsyncContextManager[CompanionInfo]:
        self._logger.debug(f"getting companion for {target_udid}")
        companion = self._get_companion_for_target(target_udid=target_udid)
        if companion:
            yield companion
        elif target_udid is None:
            raise Exception("Please provide a UDID for your target")
        elif self.companion_spawner and target_udid in self._udid_target_map:
            self._logger.debug(f"spawning a companion for {target_udid}")
            port = await self.companion_spawner.spawn_companion(
                target_udid=target_udid
            )
            # overriding host here with localhost as spawning
            # the companion only works locally
            host = "localhost"
            self._logger.info(f"companion started at {host}:{port}")
            async with self.create_companion_for_target_with_address(
                address=Address(host=host, port=port, grpc_port=port),
                metadata=metadata,
                timeout=timeout,
            ) as companion:
                yield companion
        else:
            raise Exception(f"no companion available for {target_udid}")

    async def get_stub_for_udid(self, udid: Optional[str]) -> CompanionClient:
        is_companion_available = (
            self.is_companion_available_for_target_udid(udid) if udid else False
        )
        if udid and udid in self._stub_map and udid in self._udid_companion_map:
            return CompanionClient(
                stub=self._stub_map[udid],
                is_local=self._udid_companion_map[udid].is_local,
                udid=udid,
                logger=self._logger,
                is_companion_available=is_companion_available,
            )
        else:
            async with self.create_companion_for_target_with_udid(
                target_udid=udid
            ) as companion:
                stub = self.get_stub_for_address(
                    companion.host, none_throws(companion.grpc_port)
                )
                self._stub_map[companion.udid] = stub
                return CompanionClient(
                    stub=stub,
                    is_local=companion.is_local,
                    udid=udid,
                    logger=self._logger,
                    is_companion_available=is_companion_available,
                )

    def get_stub_for_address(self, host: str, port: int) -> CompanionServiceStub:
        self._logger.debug(f"creating grpc stub for companion at {host}:{port}")
        channel = Channel(host, port, loop=asyncio.get_event_loop())
        return CompanionServiceStub(channel)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


from io import StringIO
from typing import AsyncIterator, Dict, List, Optional, Set

from idb.common.companion import CompanionClient
from idb.common.constants import TESTS_POLL_INTERVAL
from idb.common.types import TestActivity, TestRunFailureInfo, TestRunInfo
from idb.grpc.idb_pb2 import XctestRunRequest, XctestRunResponse
from idb.common.tar import untar
from logging import Logger


Mode = XctestRunRequest.Mode
Logic = XctestRunRequest.Logic
Application = XctestRunRequest.Application
UI = XctestRunRequest.UI


def _make_request(
    test_bundle_id: str,
    app_bundle_id: str,
    test_host_app_bundle_id: Optional[str],
    is_ui_test: bool,
    is_logic_test: bool,
    tests_to_run: Optional[Set[str]],
    tests_to_skip: Optional[Set[str]],
    env: Optional[Dict[str, str]],
    args: Optional[List[str]],
    result_bundle_path: Optional[str],
    timeout: Optional[int],
) -> XctestRunRequest:
    if is_logic_test:
        mode = Mode(logic=Logic())
    elif is_ui_test:
        mode = Mode(
            ui=UI(
                app_bundle_id=app_bundle_id,
                test_host_app_bundle_id=test_host_app_bundle_id,
            )
        )
    else:
        mode = Mode(application=Application(app_bundle_id=app_bundle_id))

    return XctestRunRequest(
        mode=mode,
        test_bundle_id=test_bundle_id,
        tests_to_run=list(tests_to_run or []),
        tests_to_skip=list(tests_to_skip or []),
        environment=env,
        arguments=args,
    )


async def _write_result_bundle(
    response: XctestRunResponse, output_path: str, logger: Logger
) -> None:
    payload = response.result_bundle
    if not payload:
        return
    data = payload.data
    if not len(data):
        return
    logger.info(f"Writing result bundle to {output_path}")
    await untar(data=data, output_path=output_path)
    logger.info(f"Finished writing result bundle to {output_path}")


def _make_results(response: XctestRunResponse) -> List[TestRunInfo]:
    return [
        TestRunInfo(
            bundle_name=result.bundle_name,
            class_name=result.class_name,
            method_name=result.method_name,
            logs=list(result.logs),
            duration=result.duration,
            passed=result.status == XctestRunResponse.TestRunInfo.PASSED,
            failure_info=(
                TestRunFailureInfo(
                    message=result.failure_info.failure_message,
                    file=result.failure_info.file,
                    line=result.failure_info.line,
                )
                if result.failure_info
                else None
            ),
            activityLogs=[
                TestActivity(
                    title=activity.title, duration=activity.duration, uuid=activity.uuid
                )
                for activity in result.activityLogs or []
            ],
            crashed=result.status == XctestRunResponse.TestRunInfo.CRASHED,
        )
        for result in response.results or []
    ]


async def run_xctest(
    client: CompanionClient,
    test_bundle_id: str,
    app_bundle_id: str,
    test_host_app_bundle_id: Optional[str] = None,
    is_ui_test: bool = False,
    is_logic_test: bool = False,
    tests_to_run: Optional[Set[str]] = None,
    tests_to_skip: Optional[Set[str]] = None,
    env: Optional[Dict[str, str]] = None,
    args: Optional[List[str]] = None,
    result_bundle_path: Optional[str] = None,
    idb_log_buffer: Optional[StringIO] = None,
    timeout: Optional[int] = None,
    poll_interval_sec: float = TESTS_POLL_INTERVAL,
) -> AsyncIterator[TestRunInfo]:
    async with client.stub.xctest_run.open() as stream:
        request = _make_request(
            test_bundle_id=test_bundle_id,
            app_bundle_id=app_bundle_id,
            test_host_app_bundle_id=test_host_app_bundle_id,
            is_ui_test=is_ui_test,
            is_logic_test=is_logic_test,
            tests_to_run=tests_to_run,
            tests_to_skip=tests_to_skip,
            env=env,
            args=args,
            result_bundle_path=result_bundle_path,
            timeout=timeout,
        )
        await stream.send_message(request, end=True)
        async for response in stream:
            for output in response.log_output:
                client.logger.info(output)
            if result_bundle_path:
                await _write_result_bundle(
                    response=response,
                    output_path=result_bundle_path,
                    logger=client.logger,
                )
            for result in _make_results(response):
                yield result


CLIENT_PROPERTIES = [run_xctest]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import Optional, Tuple

from idb.common.types import AccessibilityInfo
from idb.common.companion import CompanionClient
from idb.grpc.idb_pb2 import AccessibilityInfoRequest, Point


async def client(
    client: CompanionClient, point: Optional[Tuple[int, int]]
) -> AccessibilityInfo:
    grpc_point = Point(x=point[0], y=point[1]) if point is not None else None
    response = await client.stub.accessibility_info(
        AccessibilityInfoRequest(point=grpc_point)
    )
    return AccessibilityInfo(json=response.json)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.manager.companion import CompanionClient
from idb.common.types import InstalledTestInfo
from idb.grpc.idb_pb2 import XctestListBundlesRequest


async def list_xctests(client: CompanionClient) -> List[InstalledTestInfo]:
    response = await client.stub.xctest_list_bundles(XctestListBundlesRequest())
    return [
        InstalledTestInfo(
            bundle_id=bundle.bundle_id,
            name=bundle.name,
            architectures=bundle.architectures,
        )
        for bundle in response.bundles
    ]


CLIENT_PROPERTIES = [list_xctests]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio

import idb.common.plugin as plugin
from typing import List, Dict
from logging import Logger
from idb.common.types import Server
from idb.manager.companion import CompanionManager
from idb.daemon.companion_tailer import CompanionTailer
from idb.grpc.handler import GRPCHandler
from idb.grpc.server import GRPCServer
from idb.common.logging import log_call
from argparse import Namespace
from idb.common.boot_manager import BootManager


class CompositeServer(Server):
    def __init__(self, servers: List[Server], logger: Logger) -> None:
        self.servers = servers
        self.logger = logger

    def close(self) -> None:
        self.logger.info(f"Stopping {len(self.servers)} servers")
        for server in self.servers:
            self.logger.info(f"Closing {server}")
            server.close()

    async def wait_closed(self) -> None:
        await asyncio.gather(*[server.wait_closed() for server in self.servers])

    @property
    def ports(self) -> Dict[str, str]:
        return {
            key: value
            for server in self.servers
            for (key, value) in server.ports.items()
        }


@log_call()
async def start_daemon_server(args: Namespace, logger: Logger) -> Server:
    grpc_port = args.daemon_grpc_port
    notifier_path = args.notifier_path
    companion_manager = CompanionManager(companion_path=notifier_path, logger=logger)
    boot_manager = BootManager(companion_path=notifier_path)
    grpc_handler = GRPCHandler(
        companion_manager=companion_manager, boot_manager=boot_manager, logger=logger
    )
    grpc_server = GRPCServer(handler=grpc_handler, logger=logger)
    await grpc_server.start("localhost", grpc_port)
    servers: List[Server] = [grpc_server]
    if notifier_path:
        companion_tailer = CompanionTailer(
            notifier_path=notifier_path, companion_manager=companion_manager
        )
        await companion_tailer.start()
        servers.append(companion_tailer)
    servers = await plugin.resolve_servers(
        args=args,
        companion_manager=companion_manager,
        boot_manager=boot_manager,
        logger=logger,
        servers=servers,
    )
    logger.debug(f"Started servers {servers}")
    return CompositeServer(servers=servers, logger=logger)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from argparse import Namespace

from idb.cli.commands.connect import get_destination
from idb.common.types import Address
from idb.utils.testing import TestCase, ignoreTaskLeaks


@ignoreTaskLeaks
class TestParser(TestCase):
    async def test_get_destination_from_host_and_port(self) -> None:
        namespace = Namespace()
        host = "localhost"
        port = 1234
        namespace.companion = host
        namespace.port = port
        namespace.grpc_port = None
        address = get_destination(args=namespace)
        assert isinstance(address, Address)
        self.assertEqual(address.host, host)
        self.assertEqual(address.grpc_port, port)

    async def test_get_destination_from_host_and_port_and_grpc_port(self) -> None:
        namespace = Namespace()
        host = "localhost"
        port = 1234
        grpc_port = 1235
        namespace.companion = host
        namespace.port = port
        namespace.grpc_port = grpc_port
        address = get_destination(args=namespace)
        assert isinstance(address, Address)
        self.assertEqual(address.host, host)
        self.assertEqual(address.port, port)
        self.assertEqual(address.grpc_port, grpc_port)

    async def test_get_destination_from_target_udid(self) -> None:
        namespace = Namespace()
        target_udid = "SFASAF-ASFAFGE"
        namespace.companion = target_udid
        udid = get_destination(args=namespace)
        self.assertEqual(target_udid, udid)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import os
from logging import Logger
from pathlib import Path
from typing import AsyncIterator

import aiofiles
import idb.common.gzip as gzip
import idb.common.tar as tar
from grpclib.const import Status
from grpclib.exceptions import GRPCError
from idb.common.companion import CompanionClient
from idb.grpc.stream import Stream, drain_to_stream
from idb.common.xctest import xctest_paths_to_tar
from idb.grpc.idb_pb2 import InstallRequest, InstallResponse, Payload
from idb.utils.typing import none_throws


Destination = InstallRequest.Destination


async def _generate_ipa_chunks(
    ipa_path: str, logger: Logger
) -> AsyncIterator[InstallRequest]:
    logger.debug(f"Generating Chunks for .ipa {ipa_path}")
    async with aiofiles.open(ipa_path, "r+b") as file:
        while True:
            chunk = await file.read(1024)
            yield InstallRequest(payload=Payload(data=chunk))
            if not chunk:
                logger.debug(f"Finished generating .ipa chunks for {ipa_path}")
                return


async def _generate_app_chunks(
    app_path: str, logger: Logger
) -> AsyncIterator[InstallRequest]:
    logger.debug(f"Generating chunks for .app {app_path}")
    async for chunk in tar.generate_tar([app_path]):
        yield InstallRequest(payload=Payload(data=chunk))
    logger.debug(f"Finished generating .app chunks {app_path}")


async def _generate_xctest_chunks(
    path: str, logger: Logger
) -> AsyncIterator[InstallRequest]:
    logger.debug(f"Generating chunks for {path}")
    async for chunk in tar.generate_tar(xctest_paths_to_tar(path)):
        yield InstallRequest(payload=Payload(data=chunk))
    logger.debug(f"Finished generating chunks {path}")


async def _generate_dylib_chunks(
    path: str, logger: Logger
) -> AsyncIterator[InstallRequest]:
    logger.debug(f"Generating chunks for {path}")
    yield InstallRequest(name_hint=os.path.basename(path))
    async for chunk in gzip.generate_gzip(path):
        yield InstallRequest(payload=Payload(data=chunk))
    logger.debug(f"Finished generating chunks {path}")


def _generate_binary_chunks(
    path: str, destination: Destination, logger: Logger
) -> AsyncIterator[InstallRequest]:
    if destination == InstallRequest.APP:
        if path.endswith(".ipa"):
            return _generate_ipa_chunks(ipa_path=path, logger=logger)
        elif path.endswith(".app"):
            return _generate_app_chunks(app_path=path, logger=logger)
    elif destination == InstallRequest.XCTEST:
        return _generate_xctest_chunks(path=path, logger=logger)
    elif destination == InstallRequest.DYLIB:
        return _generate_dylib_chunks(path=path, logger=logger)
    raise GRPCError(
        status=Status(Status.FAILED_PRECONDITION),
        message=f"install invalid for {path} {destination}",
    )


async def _install_to_destination(
    client: CompanionClient, path: str, destination: Destination
) -> str:
    abs_path = str(Path(path).resolve(strict=True))
    async with client.stub.install.open() as stream:
        await stream.send_message(InstallRequest(destination=destination))
        await stream.send_message(InstallRequest(payload=Payload(file_path=abs_path)))
        await stream.end()
        response = await stream.recv_message()
        return response.bundle_id


async def install(client: CompanionClient, bundle_path: str) -> str:
    return await _install_to_destination(
        client=client, path=bundle_path, destination=InstallRequest.APP
    )


async def install_xctest(client: CompanionClient, bundle_path: str) -> str:
    return await _install_to_destination(
        client=client, path=bundle_path, destination=InstallRequest.XCTEST
    )


async def install_dylib(client: CompanionClient, dylib_path: str) -> str:
    return await _install_to_destination(
        client=client, path=dylib_path, destination=InstallRequest.DYLIB
    )


async def daemon(
    client: CompanionClient, stream: Stream[InstallResponse, InstallRequest]
) -> None:
    destination_message = none_throws(await stream.recv_message())
    payload_message = none_throws(await stream.recv_message())
    file_path = payload_message.payload.file_path
    destination = destination_message.destination
    async with client.stub.install.open() as forward_stream:
        await forward_stream.send_message(destination_message)
        if client.is_local:
            response = await forward_stream.send_message(payload_message)
        else:
            response = await drain_to_stream(
                stream=forward_stream,
                generator=_generate_binary_chunks(
                    path=file_path, destination=destination, logger=client.logger
                ),
                logger=client.logger,
            )
        await stream.send_message(response, end=True)


CLIENT_PROPERTIES = [install, install_xctest, install_dylib]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import json
import logging
import os
import signal
from typing import List

from idb.common.constants import IDB_DAEMON_PID_PATH


def save_daemon_pid(pid: int) -> None:
    pids = _get_daemon_pids()
    pids.append(pid)
    _write_daemon_pids(pids=pids)
    logging.debug(f"saved daemon pid {pid}")


def remove_daemon_pid(pid: int) -> None:
    pids = _get_daemon_pids()
    if pids.count(pid) > 0:
        pids.remove(pid)
        _write_daemon_pids(pids=pids)
        logging.debug(f"removed daemon pid {pid}")


def _write_daemon_pids(pids: List[int]) -> None:
    with open(IDB_DAEMON_PID_PATH, "w") as pid_file:
        json.dump(pids, pid_file)
        pid_file.flush()


def _has_saved_pids() -> bool:
    pids = _get_daemon_pids()
    logging.debug(f"has saved pids {pids}")
    return len(pids) > 0


def _get_daemon_pids() -> List[int]:
    try:
        with open(IDB_DAEMON_PID_PATH) as pid_file:
            return json.load(pid_file)
    except Exception:
        return []


def _clear_saved_daemon_pids() -> None:
    if os.path.exists(IDB_DAEMON_PID_PATH):
        # Empty the file
        with open(IDB_DAEMON_PID_PATH, "wb", buffering=0) as pid_file:
            pid_file.flush()


async def kill_saved_pids() -> None:
    if not _has_saved_pids():
        logging.debug(f"no daemon pid found")
        return
    for pid in _get_daemon_pids():
        try:
            os.kill(pid, signal.SIGTERM)
            logging.info(f"stopped daemon with pid {pid}")
        except OSError:
            logging.exception(f"failed to stop daemon with pid {pid} because of error")
    _clear_saved_daemon_pids()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


from idb.common.companion import CompanionClient
from idb.grpc.idb_pb2 import TerminateRequest


async def client(client: CompanionClient, bundle_id: str) -> None:
    await client.stub.terminate(TerminateRequest(bundle_id=bundle_id))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.common.companion import CompanionClient
from idb.grpc.stream import Stream, drain_to_stream
from idb.common.stream import stream_map
from idb.common.tar import generate_tar
from idb.grpc.idb_pb2 import AddMediaRequest, AddMediaResponse, Payload


async def client(client: CompanionClient, file_paths: List[str]) -> None:
    async with client.stub.add_media.open() as stream:
        for file_path in file_paths:
            await stream.send_message(
                AddMediaRequest(payload=Payload(file_path=file_path))
            )
        await stream.end()
        await stream.recv_message()


async def daemon(
    client: CompanionClient, stream: Stream[AddMediaResponse, AddMediaRequest]
) -> None:
    async with client.stub.add_media.open() as companion:
        if client.is_local:
            generator = stream
        else:
            paths = [request.payload.file_path async for request in stream]
            generator = stream_map(
                generate_tar(paths=paths, place_in_subfolders=True),
                lambda chunk: AddMediaRequest(payload=Payload(data=chunk)),
            )
        response = await drain_to_stream(
            stream=companion, generator=generator, logger=client.logger
        )
        await stream.send_message(response)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from argparse import Namespace

from idb.cli.commands.base import BaseCommand
from idb.client.client import IdbClient


class KillCommand(BaseCommand):
    @property
    def description(self) -> str:
        return "Kill the idb daemon"

    @property
    def name(self) -> str:
        return "kill"

    async def _run_impl(self, args: Namespace) -> None:
        await IdbClient.kill()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from grpclib.const import Status
from grpclib.exceptions import GRPCError
from idb.common.companion import CompanionClient
from idb.grpc.ipc_loader import DaemonContext
from idb.grpc.idb_pb2 import BootRequest, BootResponse
from idb.utils.typing import none_throws


async def client(client: CompanionClient) -> None:
    await client.stub.boot(BootRequest())


async def daemon(
    context: DaemonContext, client: CompanionClient, request: BootRequest
) -> BootResponse:
    if client.is_companion_available:
        await none_throws(context.boot_manager).boot(udid=none_throws(client.udid))
        return BootResponse()
    else:
        # TODO T41660845
        raise GRPCError(
            status=Status(Status.UNIMPLEMENTED),
            message="boot with chained daemons hasn't been implemented yet",
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import argparse
import asyncio
import concurrent.futures
import logging
import sys
from typing import List, Optional, Set

from idb.cli.commands.accessibility import (
    AccessibilityInfoAllCommand,
    AccessibilityInfoAtPointCommand,
)
from idb.cli.commands.add_media import AddMediaCommand
from idb.cli.commands.approve import ApproveCommand
from idb.cli.commands.base import Command, CommandGroup
from idb.cli.commands.boot import BootCommand
from idb.cli.commands.clear_keychain import ClearKeychainCommand
from idb.cli.commands.connect import ConnectCommand, ConnectCommandException
from idb.cli.commands.contacts import ContactsUpdateCommand
from idb.cli.commands.crash import (
    CrashDeleteCommand,
    CrashListCommand,
    CrashShowCommand,
)
from idb.cli.commands.daemon import DaemonCommand
from idb.cli.commands.debugserver import (
    DebugServerStartCommand,
    DebugServerStatusCommand,
    DebugServerStopCommand,
)
from idb.cli.commands.describe import DescribeCommand
from idb.cli.commands.disconnect import DisconnectCommand
from idb.cli.commands.file import (
    DeprecatedPullCommand,
    DeprecatedPushCommand,
    FSListCommand,
    FSMkdirCommand,
    FSMoveCommand,
    FSPullCommand,
    FSPushCommand,
    FSRemoveCommand,
)
from idb.cli.commands.focus import FocusCommand
from idb.cli.commands.hid import (
    ButtonCommand,
    KeyCommand,
    KeySequenceCommand,
    SwipeCommand,
    TapCommand,
    TextCommand,
)
import idb.common.plugin as plugin
from idb.cli.commands.install import InstallAppCommand, InstallDylibCommand
from idb.cli.commands.instruments import InstrumentsCommand
from idb.cli.commands.kill import KillCommand
from idb.cli.commands.launch import LaunchCommand
from idb.cli.commands.list_apps import ListAppsCommand
from idb.cli.commands.list_targets import ListTargetsCommand
from idb.cli.commands.log import LogCommand
from idb.cli.commands.open_url import OpenUrlCommand
from idb.cli.commands.record import RecordVideoCommand
from idb.cli.commands.screenshot import ScreenshotCommand
from idb.cli.commands.set_location import SetLocationCommand
from idb.cli.commands.terminate import TerminateCommand
from idb.cli.commands.uninstall import UninstallCommand
from idb.cli.commands.xctest_install import InstallXctestCommand
from idb.cli.commands.xctest_list import ListXctestsCommand
from idb.cli.commands.xctest_list_bundle import ListTestBundleCommand
from idb.cli.commands.xctest_run import RunXctestCommand
from idb.common.types import IdbException


COROUTINE_DRAIN_TIMEOUT = 2


# Set the logger's basicConfig first, then add the logview handlers
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s [%(levelname)s] - %(name)s - %(message)s"
)
logger: logging.Logger = logging.getLogger()


async def gen_main(cmd_input: Optional[List[str]] = None,) -> int:
    # Setup parser
    parser = argparse.ArgumentParser(
        description="idb cli: versatile tool "
        "to communicate with an iOS target (simulator/device)",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    commands: List[Command] = [
        DescribeCommand(),
        InstallAppCommand(),
        InstallDylibCommand(),
        UninstallCommand(),
        ListAppsCommand(),
        LaunchCommand(),
        TerminateCommand(),
        CommandGroup(
            name="xctest",
            description="Operations with xctest on target",
            commands=[
                InstallXctestCommand(),
                ListXctestsCommand(),
                ListTestBundleCommand(),
                RunXctestCommand(),
            ],
        ),
        CommandGroup(
            name="file",
            description="File operations on target",
            commands=[
                FSMoveCommand(),
                FSPullCommand(),
                FSPushCommand(),
                FSMkdirCommand(),
                FSRemoveCommand(),
                FSListCommand(),
            ],
        ),
        CommandGroup(
            name="contacts",
            description="Contacts database operations on target",
            commands=[ContactsUpdateCommand()],
        ),
        LogCommand(),
        CommandGroup(
            name="record",
            description="Record what the screen is doing",
            commands=[RecordVideoCommand()],
        ),
        RecordVideoCommand(),
        DeprecatedPushCommand(),
        DeprecatedPullCommand(),
        OpenUrlCommand(),
        ClearKeychainCommand(),
        SetLocationCommand(),
        ApproveCommand(),
        ConnectCommand(),
        DisconnectCommand(),
        ListTargetsCommand(),
        DaemonCommand(),
        ScreenshotCommand(),
        CommandGroup(
            name="ui",
            description="UI interactions on target",
            commands=[
                AccessibilityInfoAllCommand(),
                AccessibilityInfoAtPointCommand(),
                TapCommand(),
                ButtonCommand(),
                TextCommand(),
                KeyCommand(),
                KeySequenceCommand(),
                SwipeCommand(),
            ],
        ),
        CommandGroup(
            name="crash",
            description="Operations on crashes",
            commands=[CrashListCommand(), CrashShowCommand(), CrashDeleteCommand()],
        ),
        InstrumentsCommand(),
        KillCommand(),
        AddMediaCommand(),
        FocusCommand(),
        BootCommand(),
        CommandGroup(
            name="debugserver",
            description="debugserver interactions",
            commands=[
                DebugServerStartCommand(),
                DebugServerStopCommand(),
                DebugServerStatusCommand(),
            ],
        ),
    ]
    sorted_commands = sorted(commands, key=lambda command: command.name)
    root_command = CommandGroup("root_command", "", sorted_commands)
    root_command.add_parser_arguments(parser)

    # Parse input and run
    cmd_input = cmd_input or sys.argv[1:]

    try:
        args = parser.parse_args(cmd_input)
        plugin.on_launch(logger)
        await root_command.run(args, argparse.Namespace())
        return 0
    except ConnectCommandException as e:
        print(str(e))
        return 1
    except IdbException as e:
        logger.critical(e.args[0])
        return 1
    except Exception:
        logger.exception("Exception thrown in main")
        return 1
    finally:
        await plugin.on_close(logger)
        pending = set(asyncio.Task.all_tasks())
        pending.discard(asyncio.Task.current_task())
        await drain_coroutines(pending)


async def drain_coroutines(pending: Set[asyncio.Task]) -> None:
    if not pending:
        return
    logger.debug(f"Shutting down {len(pending)} coroutines")
    try:
        await asyncio.wait_for(
            asyncio.shield(asyncio.gather(*pending)), timeout=COROUTINE_DRAIN_TIMEOUT
        )
        logger.debug(f"Drained all coroutines")
    except asyncio.TimeoutError:
        logger.debug(f"Timeout waiting for coroutines to drain")
    except concurrent.futures.CancelledError:
        pass


def main(cmd_input: Optional[List[str]] = None,) -> int:
    loop = asyncio.get_event_loop()
    try:
        return loop.run_until_complete(gen_main(cmd_input))
    finally:
        loop.close()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.common.types import CrashLogInfo, CrashLogQuery
from idb.common.companion import CompanionClient
from idb.ipc.mapping.crash import _to_crash_log_info_list, _to_crash_log_query_proto


async def client(client: CompanionClient, query: CrashLogQuery) -> List[CrashLogInfo]:
    response = await client.stub.crash_list(_to_crash_log_query_proto(query))
    return _to_crash_log_info_list(response)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from logging import Logger
from typing import Any, AsyncIterator, Dict, Generic, Optional, TypeVar

from idb.utils.typing import none_throws


_TSend = TypeVar("_TSend")
_TRecv = TypeVar("_TRecv")


class Stream(Generic[_TSend, _TRecv], AsyncIterator[_TRecv]):
    metadata: Dict[str, str]  # pyre-ignore

    async def recv_message(self) -> Optional[_TRecv]:
        ...

    async def send_message(self, message: _TSend, end: bool = False) -> None:
        ...

    async def end(self) -> None:
        ...


async def drain_to_stream(
    stream: Stream[_TSend, _TRecv], generator: AsyncIterator[_TSend], logger: Logger
) -> _TRecv:
    while True:
        async for message in generator:
            await stream.send_message(message)
        await stream.end()
        logger.debug(f"Streamed all chunks to companion, waiting for completion")
        response = none_throws(await stream.recv_message())
        logger.debug(f"Companion completed")
        return response


async def generate_bytes(
    stream: AsyncIterator[Any],  # pyre-ignore
) -> AsyncIterator[bytes]:
    async for response in stream:
        yield response.payload.data
